"""COMPASS Ordinance Location Validation logic

These are primarily used to validate that a legal document applies to a
particular location.
"""

import asyncio
import logging

from compass.llm.calling import BaseLLMCaller, ChatLLMCaller
from compass.common import setup_async_decision_tree, run_async_tree
from compass.validation.graphs import (
    setup_graph_correct_jurisdiction_type,
    setup_graph_correct_jurisdiction_from_url,
)
from compass.utilities.enums import LLMUsageCategory


logger = logging.getLogger(__name__)


class DTreeURLCountyValidator(BaseLLMCaller):
    """Validator that checks whether a URL matches a jurisdiction"""

    SYSTEM_MESSAGE = (
        "You are an expert data analyst that examines URLs to determine if "
        "they contain information about jurisdictions. only ever answer "
        "based on the information in the URL itself."
    )

    def __init__(self, jurisdiction, **kwargs):
        """

        Parameters
        ----------
        structured_llm_caller : `StructuredLLMCaller`
            StructuredLLMCaller instance. Used for structured validation
            queries.
        """
        super().__init__(**kwargs)
        self.jurisdiction = jurisdiction

    async def check(self, url):
        """Check if the content passes the validation

        Parameters
        ----------
        content : str
            Document content to validate.

        Returns
        -------
        bool
            ``True`` if the content passes the validation check,
            ``False`` otherwise.
        """
        if not url:
            return False

        chat_llm_caller = ChatLLMCaller(
            llm_service=self.llm_service,
            system_message=self.SYSTEM_MESSAGE,
            usage_tracker=self.usage_tracker,
            **self.kwargs,
        )
        tree = setup_async_decision_tree(
            setup_graph_correct_jurisdiction_from_url,
            usage_sub_label=LLMUsageCategory.DOCUMENT_JURISDICTION_VALIDATION,
            jurisdiction=self.jurisdiction,
            url=url,
            chat_llm_caller=chat_llm_caller,
        )
        out = await run_async_tree(tree, response_as_json=True)
        return self._parse_output(out)

    def _parse_output(self, props):  # noqa: PLR6301
        """Parse LLM response and return boolean validation result"""
        logger.debug(
            "Parsing URL jurisdiction validation output:\n\t%s", props
        )
        return len(props) > 0 and all(props.values())


class DTreeJurisdictionValidator(BaseLLMCaller):
    """Jurisdiction Validation using a decision tree"""

    META_SCORE_KEY = "Jurisdiction Validation Score"
    SYSTEM_MESSAGE = (
        "You are a legal expert assisting a user with determining the scope "
        "of applicability for their legal ordinance documents."
    )

    def __init__(self, jurisdiction, **kwargs):
        """

        Parameters
        ----------
        structured_llm_caller : `StructuredLLMCaller`
            StructuredLLMCaller instance. Used for structured validation
            queries.
        """
        super().__init__(**kwargs)
        self.jurisdiction = jurisdiction

    async def check(self, content):
        """Check if the content passes the validation

        Parameters
        ----------
        content : str
            Document content to validate.

        Returns
        -------
        bool
            ``True`` if the content passes the validation check,
            ``False`` otherwise.
        """
        if not content:
            return False

        chat_llm_caller = ChatLLMCaller(
            llm_service=self.llm_service,
            system_message=self.SYSTEM_MESSAGE,
            usage_tracker=self.usage_tracker,
            **self.kwargs,
        )
        tree = setup_async_decision_tree(
            setup_graph_correct_jurisdiction_type,
            usage_sub_label=LLMUsageCategory.DOCUMENT_JURISDICTION_VALIDATION,
            jurisdiction=self.jurisdiction,
            text=content,
            chat_llm_caller=chat_llm_caller,
        )
        out = await run_async_tree(tree, response_as_json=True)
        return self._parse_output(out)

    def _parse_output(self, props):  # noqa: PLR6301
        """Parse LLM response and return boolean validation result"""
        logger.debug(
            "Parsing county jurisdiction validation output:\n\t%s", props
        )
        return props.get("correct_jurisdiction")


class JurisdictionValidator:
    """COMPASS Ordinance Jurisdiction validator

    Combines the logic of several validators into a single class.

    Purpose:
        Determine whether a document pertains to a specific county.
    Responsibilities:
        1. Use a combination of heuristics and LLM queries to determine
           whether or not a document pertains to a particular county.
    Key Relationships:
        Uses a :class:`~compass.llm.calling.StructuredLLMCaller` for
        LLM queries and delegates sub-validation to
        :class:`DTreeJurisdictionValidator`,
        and :class:`DTreeURLCountyValidator`.
    """

    def __init__(self, score_thresh=0.8, text_splitter=None, **kwargs):
        """

        Parameters
        ----------
        score_thresh : float, optional
            Score threshold to exceed when voting on content from raw
            pages. By default, ``0.8``.
        text_splitter : langchain.text_splitter.TextSplitter, optional
            Optional text splitter instance to attach to doc (used for
            splitting out pages in an HTML document).
            By default, ``None``.
        """
        self.score_thresh = score_thresh
        self.text_splitter = text_splitter
        self.kwargs = kwargs

    async def check(self, doc, jurisdiction):
        """Check if the document belongs to the county

        Parameters
        ----------
        doc : :class:`elm.web.document.BaseDocument`
            Document instance. Should contain a "source" key in the
            ``attrs`` that contains a URL (used for the URL validation
            check). Raw content will be parsed for county name and
            correct jurisdiction.

        Returns
        -------
        bool
            `True` if the doc contents pertain to the input county.
            `False` otherwise.
        """
        if hasattr(doc, "text_splitter") and self.text_splitter is not None:
            old_splitter = doc.text_splitter
            doc.text_splitter = self.text_splitter
            out = await self._check(doc, jurisdiction)
            doc.text_splitter = old_splitter
            return out

        return await self._check(doc, jurisdiction)

    async def _check(self, doc, jurisdiction):
        """Check if the document belongs to the county"""
        if self.text_splitter is not None:
            doc.text_splitter = self.text_splitter

        url = doc.attrs.get("source")
        if url:
            logger.debug("Checking URL (%s) for county name...", url)
            url_validator = DTreeURLCountyValidator(
                jurisdiction, **self.kwargs
            )
            url_is_correct_jurisdiction = await url_validator.check(url)
            if url_is_correct_jurisdiction:
                return True

        logger.info("Validating document from source: %s", url or "Unknown")
        logger.debug("Checking for correct for jurisdiction...")
        jurisdiction_validator = DTreeJurisdictionValidator(
            jurisdiction, **self.kwargs
        )
        return await _validator_check_for_doc(
            validator=jurisdiction_validator,
            doc=doc,
            score_thresh=self.score_thresh,
        )


async def _validator_check_for_doc(validator, doc, score_thresh=0.9, **kwargs):
    """Apply a validator check to a doc's raw pages"""
    outer_task_name = asyncio.current_task().get_name()
    validation_checks = [
        asyncio.create_task(
            validator.check(text, **kwargs), name=outer_task_name
        )
        for text in doc.raw_pages
    ]
    out = await asyncio.gather(*validation_checks)
    score = _weighted_vote(out, doc)
    doc.attrs[validator.META_SCORE_KEY] = score
    logger.debug(
        "%s is %.2f for doc from source %s (Pass: %s; threshold: %.2f)",
        validator.META_SCORE_KEY,
        score,
        doc.attrs.get("source", "Unknown"),
        str(score >= score_thresh),
        score_thresh,
    )
    return score >= score_thresh


def _weighted_vote(out, doc):
    """Compute weighted average of responses based on text length"""
    if not doc.raw_pages:
        return 0

    total = weights = 0
    for verdict, text in zip(out, doc.raw_pages, strict=True):
        if verdict is None:
            continue
        weight = len(text)
        logger.debug("Weight=%d, Verdict=%d", weight, int(verdict))
        weights += weight
        total += verdict * weight

    weights = max(weights, 1)
    return total / weights



"""Document validation decision tree graph setup functions"""

from compass.common import (
    setup_graph_no_nodes,
    llm_response_starts_with_yes,
    llm_response_starts_with_no,
)


def setup_graph_correct_document_type(**kwargs):
    """Setup graph to check for correct document type in legal text

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = setup_graph_no_nodes(**kwargs)  # noqa: N806
    G.add_node(
        "init",
        prompt=(
            "Does the following text excerpt detail in-effect legal statutes? "
            "Please start your response with either 'Yes' or 'No' and "
            "briefly explain your answer."
            '\n\n"""\n{text}\n"""'
        ),
    )

    G.add_edge("init", "is_model", condition=llm_response_starts_with_yes)
    G.add_node(
        "is_model",
        prompt=(
            "Is the text an excerpt from a model ordinance or other "
            "kind of model law? "
            "Please start your response with either 'Yes' or 'No' and briefly "
            "explain why you chose your answer."
        ),
    )

    G.add_edge("is_model", "is_pd", condition=llm_response_starts_with_no)
    G.add_node(
        "is_pd",
        prompt=(
            "Is the text an excerpt from a planning document? "
            "Please start your response with either 'Yes' or 'No' and briefly "
            "explain why you chose your answer."
        ),
    )

    G.add_edge("is_pd", "is_pres", condition=llm_response_starts_with_no)
    G.add_node(
        "is_pres",
        prompt=(
            "Is the text an excerpt from a presentation? "
            "Please start your response with either 'Yes' or 'No' and briefly "
            "explain why you chose your answer."
        ),
    )

    G.add_edge("is_pres", "is_draft", condition=llm_response_starts_with_no)
    G.add_node(
        "is_draft",
        prompt=(
            "Is the text an excerpt from a draft document? "
            "Please start your response with either 'Yes' or 'No' and briefly "
            "explain why you chose your answer."
        ),
    )

    G.add_edge("is_draft", "is_report", condition=llm_response_starts_with_no)
    G.add_node(
        "is_report",
        prompt=(
            "Is the text an excerpt from a report or summary "
            "document? "
            "Please start your response with either 'Yes' or 'No' and briefly "
            "explain why you chose your answer."
        ),
    )

    G.add_edge(
        "is_report", "is_article", condition=llm_response_starts_with_no
    )
    G.add_node(
        "is_article",
        prompt=(
            "Is the text an excerpt from a news article or "
            "other media? "
            "Please start your response with either 'Yes' or 'No' and briefly "
            "explain why you chose your answer."
        ),
    )

    G.add_edge("is_article", "final", condition=llm_response_starts_with_no)
    G.add_node(
        "final",
        prompt=(
            "Respond based on our entire conversation so far. Return your "
            "answer as a dictionary in JSON format (not markdown). Your JSON "
            "file must include exactly three keys:\n\n"
            "1. **'summary'** (string) - A concise summary of the text.\n"
            "2. **'type'** (string) - The best-fitting category for the "
            "source of the text.\n"
            "3. **'{key}'** (boolean) -\n"
            "\t- `true` if the text is a **legally binding regulation**.\n"
            "\t- `false` if the text belongs to any other type of document or "
            "if you cannot tell for certain one way or another.\n\n"
        ),
    )
    return G


def setup_graph_correct_jurisdiction_type(jurisdiction, **kwargs):
    """Setup graph to check for correct jurisdiction type in legal text

    Parameters
    ----------
    jurisdiction : compass.utilities.location.Jurisdiction
        Jurisdiction for which validation is being performed.
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = setup_graph_no_nodes(**kwargs)  # noqa: N806

    G.add_node(
        "init",
        prompt=(
            "Does the following legal text explicitly include enough "
            "information to reasonably conclude what type of "
            "jurisdiction it applies to? Common types of jurisdictions "
            "include 'state', 'county', 'city', 'township',' borough', etc. "
            "Please start your response with either 'Yes' or 'No' and briefly "
            "explain your answer."
            '\n\n"""\n{text}\n"""'
        ),
    )

    G.add_edge("init", "is_state", condition=llm_response_starts_with_yes)
    G.add_node(
        "is_state",
        prompt=(
            "Based on the legal text, is it reasonable to conclude that the "
            "provisions within apply to the entire state of "
            f"{jurisdiction.state}, either directly or through reference to "
            "a statewide statute, agency, or regulatory authority? If the "
            "text only applies to a county, municipality, or other local "
            f"subdivision within {jurisdiction.state}, or if there is no "
            "reasonable basis to infer statewide application, respond with "
            "'No'. Start your response with 'Yes' or 'No' and briefly explain."
        ),
    )
    node_to_connect = "is_state"

    if jurisdiction.county:
        G.add_edge(
            node_to_connect, "is_county", condition=llm_response_starts_with_no
        )
        G.add_edge(
            node_to_connect, "final", condition=llm_response_starts_with_yes
        )
        G.add_node(
            "is_county",
            prompt=(
                "Based on the legal text, is it reasonable to conclude that "
                "the provisions within apply to the entire area governed by "
                f"{jurisdiction.full_county_phrase} (including unincorporated "
                "areas), either directly or through reference to a "
                "county-wide code, planning authority, commission, or joint "
                "resolution with other local governments? If the provisions "
                "within the text apply **only** to a subdivision of "
                f"{jurisdiction.full_county_phrase} (such as a city or "
                "township), or the scope is unclear, respond with 'No'. "
                "Start your answer with 'Yes' or 'No' and explain briefly."
            ),
        )
        G.add_edge(
            "is_county", "final", condition=llm_response_starts_with_yes
        )
        node_to_connect = "is_county"

    if jurisdiction.subdivision_name:
        G.add_edge(
            node_to_connect, "is_city", condition=llm_response_starts_with_no
        )
        G.add_edge(
            node_to_connect, "final", condition=llm_response_starts_with_yes
        )
        G.add_node(
            "is_city",
            prompt=(
                "Based on the legal text, is it reasonable to conclude that "
                "the provisions apply specifically to "
                f"{jurisdiction.full_subdivision_phrase} (rather than a "
                "county, state, or federal jurisdiction)? If the text instead "
                "applies to a broader jurisdiction, or does not provide a "
                "reasonable basis to infer that it is limited to municipal "
                "governance, respond with 'No'. "
                "Start your response with 'Yes' or 'No' and briefly explain."
            ),
        )
        node_to_connect = "is_city"

    G.add_edge(node_to_connect, "final")
    G.add_node(
        "final",
        prompt=(
            "Respond based on our entire conversation so far. Return your "
            "answer as a dictionary in JSON format (not markdown). Your JSON "
            "file must include exactly two keys. The keys are "
            "'correct_jurisdiction' and 'explanation'. The value of the "
            "'correct_jurisdiction' key should be a boolean that is set to "
            "`true` **only if** it is reasonable to conclude that the "
            "provisions within apply to the entire area (i.e. "
            f"{jurisdiction.type}-wide) governed by {jurisdiction.full_name} "
            "(`false` otherwise). The value of the 'explanation' key should "
            "be a string containing a brief explanation for your choice. "
        ),
    )
    return G


def setup_graph_correct_jurisdiction_from_url(jurisdiction, **kwargs):
    """Setup graph to check for correct jurisdiction in URL

    Parameters
    ----------
    jurisdiction : compass.utilities.location.Jurisdiction
        Jurisdiction for which validation is being performed.
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """

    G = setup_graph_no_nodes(**kwargs)  # noqa: N806

    G.add_node(
        "init",
        prompt=(
            f"Does the following URL explicitly mention {jurisdiction.state} "
            "state in some way (e.g. either by full name or abbreviation)? "
            "**Do not** answer based on auxiliary information like county or "
            "city names. "
            "Please start your response with either 'Yes' or 'No' and explain "
            "your answer."
            "\n\nURL: '{url}\n'"
        ),
    )

    node_to_connect = "init"
    keys_to_collect = {"correct_state": f"{jurisdiction.state} state"}

    if jurisdiction.county:
        G.add_edge(
            node_to_connect,
            "mentions_county",
            condition=llm_response_starts_with_yes,
        )
        G.add_node(
            "mentions_county",
            prompt=(
                "Does the URL explicitly mention "
                f"{jurisdiction.full_county_phrase} in some way (e.g. either "
                "by full name or abbreviation)? **Do not** answer based on "
                "auxiliary information like state or city names. "
                "Please start your response with either 'Yes' or 'No' and "
                "briefly explain your answer."
                "\n\nURL: '{url}\n'"
            ),
        )
        keys_to_collect["correct_county"] = (
            f"{jurisdiction.full_county_phrase}"
        )
        node_to_connect = "mentions_county"

    if jurisdiction.subdivision_name:
        G.add_edge(
            node_to_connect,
            "mentions_city",
            condition=llm_response_starts_with_yes,
        )
        G.add_node(
            "mentions_city",
            prompt=(
                "Does the URL explicitly mention "
                f"{jurisdiction.full_subdivision_phrase} in some way (e.g. "
                "either by full name or abbreviation)? **Do not** answer "
                "based on auxiliary information like state or county names. "
                "Please start your response with either 'Yes' or 'No' and "
                "briefly explain your answer."
                "\n\nURL: '{url}\n'"
            ),
        )
        keys_to_collect[f"correct_{jurisdiction.type.casefold()}"] = (
            f"{jurisdiction.full_subdivision_phrase}"
        )
        node_to_connect = "mentions_city"

    G.add_edge(
        node_to_connect, "final", condition=llm_response_starts_with_yes
    )
    G.add_node("final", prompt=_compile_final_url_prompt(keys_to_collect))
    return G


def _compile_final_url_prompt(keys_to_collect):
    """Compile final URL instruction prompt"""
    num_keys = len(keys_to_collect) + 1
    num_keys = f"Your JSON file must include exactly {num_keys} keys. "

    out_keys = ", ".join([f"'{key}'" for key in keys_to_collect])
    out_keys = f"The keys are {out_keys} and 'explanation'. "

    explain_text = _compile_url_key_explain_text(keys_to_collect)

    return (
        "Respond based on our entire conversation so far. Return your "
        "answer as a dictionary in JSON format (not markdown). "
        f"{num_keys}{out_keys}{explain_text}"
    )


def _compile_url_key_explain_text(keys_to_collect):
    """Compile explanations ofr each output key"""
    explain_text = []
    for key, name in keys_to_collect.items():
        explain_text.append(
            f"The value of the '{key}' key should be a boolean that is set to "
            f"`True` if the URL explicitly mentions {name} in some way "
            "(`False` otherwise). "
        )

    choices = "choices" if len(keys_to_collect) > 1 else "choice"
    explain_text.append(
        "The value of the 'explanation' key should be a string containing a "
        f"short explanation for your {choices}. "
    )
    return "".join(explain_text)



"""Common ordinance extraction components"""

import asyncio
import logging
from datetime import datetime

import networkx as nx
from elm import ApiBase

from compass.common.tree import AsyncDecisionTree
from compass.utilities import llm_response_as_json
from compass.utilities.enums import LLMUsageCategory
from compass.utilities.parsing import (
    merge_overlapping_texts,
    clean_backticks_from_llm_response,
)
from compass.exceptions import COMPASSRuntimeError


logger = logging.getLogger(__name__)
_SECTION_PROMPT = (
    "The value of the 'section' key should be a string representing the "
    "title of the section (including numerical labels), if it's given, "
    "and `null` otherwise."
)
_SUMMARY_PROMPT = (
    "The value of the 'summary' key should be a short summary "
    "of the ordinance, using direct text excerpts as much as possible."
)
_UNITS_IN_SUMMARY_PROMPT = (
    "Include any clarifications about the units in the summary."
)
EXTRACT_ORIGINAL_TEXT_PROMPT = (
    "Extract all portions of the text (with original formatting) "
    "that state how close I can site {tech} to {feature}. "
    "Focus on ordinances relating to setbacks from {feature}; do not "
    "respond based on any text related to {ignore_features}. "
    "Please only consider ordinances for systems that would typically be "
    "defined as {tech} based on the text itself — for example, systems "
    "intended for electricity generation or sale, or those above "
    "thresholds such as height or rated capacity. Ignore any text that "
    "applies **only** to smaller or clearly non-commercial systems. "
)


def setup_graph_no_nodes(**kwargs):
    """Setup a graph with no nodes

    This function is used to set keywords on the graph that can be used
    in text prompts on the graph nodes.

    Returns
    -------
    nx.DiGraph
        Graph with no nodes but with global keywords set.
    """
    return nx.DiGraph(
        SECTION_PROMPT=_SECTION_PROMPT,
        SUMMARY_PROMPT=_SUMMARY_PROMPT,
        UNITS_IN_SUMMARY_PROMPT=_UNITS_IN_SUMMARY_PROMPT,
        **kwargs,
    )


def llm_response_starts_with_yes(response):
    """Check if LLM response begins with "yes" (case-insensitive)

    Parameters
    ----------
    response : str
        LLM response string.

    Returns
    -------
    bool
        `True` if LLM response begins with "Yes".
    """
    return response.lower().startswith("yes")


def llm_response_starts_with_no(response):
    """Check if LLM response begins with "no" (case-insensitive)

    Parameters
    ----------
    response : str
        LLM response string.

    Returns
    -------
    bool
        `True` if LLM response begins with "No".
    """
    return response.lower().startswith("no")


def llm_response_does_not_start_with_no(response):
    """Check if LLM response does not start with "no" (case-insensitive)

    Parameters
    ----------
    response : str
        LLM response string.

    Returns
    -------
    bool
        `True` if LLM response does not begin with "No".
    """
    return not llm_response_starts_with_no(response)


def setup_async_decision_tree(
    graph_setup_func, usage_sub_label=None, **kwargs
):
    """Setup Async Decision tree for ordinance extraction"""
    G = graph_setup_func(**kwargs)  # noqa: N806
    tree = AsyncDecisionTree(G, usage_sub_label=usage_sub_label)
    assert len(tree.chat_llm_caller.messages) == 1
    return tree


def found_ord(messages):
    """Check messages from the base graph to see if ordinance was found

    ..IMPORTANT:: This function may break if the base graph structure
                  changes. Always update the hardcoded values to match
                  the base graph message containing the LLM response
                  about ordinance content.
    """
    num_messages_in_base_tree = 3
    if len(messages) < num_messages_in_base_tree:
        return False
    return llm_response_starts_with_yes(messages[2].get("content", ""))


async def run_async_tree(tree, response_as_json=True):
    """Run Async Decision Tree and return output as dict"""
    try:
        response = await tree.async_run()
    except COMPASSRuntimeError:
        response = None

    if response_as_json:
        return llm_response_as_json(response) if response else {}

    return response


async def run_async_tree_with_bm(tree, base_messages):
    """Run Async Decision Tree from base messages; return dict output"""
    tree.chat_llm_caller.messages = base_messages
    assert len(tree.chat_llm_caller.messages) == len(base_messages)
    return await run_async_tree(tree)


def empty_output(feature):
    """Empty output for a feature (not found in text)"""
    if feature in {"structures", "property line"}:
        return [
            {"feature": f"{feature} (participating)"},
            {"feature": f"{feature} (non-participating)"},
        ]
    return [{"feature": feature}]


def setup_base_setback_graph(**kwargs):
    """Setup graph to get setback ordinance text for a given feature

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = setup_graph_no_nodes(**kwargs)  # noqa: N806

    G.add_node(
        "init",
        prompt=(
            "Is there text in the following legal document that describes "
            "how far I have to setback {tech} from {feature}? "
            "{feature_clarifications}"  # expected to end in space
            "Focus only on setbacks from {feature}; do not respond "
            "based on any text related to {ignore_features}. "
            "Please only consider setbacks specifically for systems that "
            "would typically be defined as {tech} based on the text itself "
            "— for example, systems intended for electricity generation or "
            "sale, or those above thresholds such as height or rated "
            "capacity. Ignore any requirements that apply only to smaller "
            "or clearly non-commercial systems. "
            "Don't forget to pay extra attention to clarifying text found "
            "in parentheses and footnotes. "
            "Please start your response with either 'Yes' or 'No' and briefly "
            "explain your answer."
            '\n\n"""\n{text}\n"""'
        ),
    )

    G.add_edge(
        "init", "get_text", condition=llm_response_does_not_start_with_no
    )
    G.add_node("get_text", prompt=EXTRACT_ORIGINAL_TEXT_PROMPT)

    return G


def setup_participating_owner(**kwargs):
    """Setup graph to check for "participating" setbacks for a feature

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = setup_graph_no_nodes(**kwargs)  # noqa: N806

    G.add_node(
        "init",
        prompt=(
            "Does the ordinance for {feature} setbacks explicitly specify "
            "a value that applies to participating property owners? "
            "Focus only on setbacks from {feature}; do not respond "
            "based on any text related to {ignore_features}. "
            "Please only consider setbacks specifically for systems that "
            "would typically be defined as {tech} based on the text itself "
            "— for example, systems intended for electricity generation or "
            "sale, or those above thresholds such as height or rated "
            "capacity. Ignore any requirements that apply only to smaller "
            "or clearly non-commercial systems. "
            "If your answer is 'yes', justify it by quoting the raw text "
            "directly."
        ),
    )
    G.add_edge("init", "non_part")
    G.add_node(
        "non_part",
        prompt=(
            "Does the ordinance for {feature} setbacks explicitly specify "
            "a value that applies to non-participating property owners? "
            "Focus only on setbacks from {feature}; do not respond "
            "based on any text related to {ignore_features}. "
            "Please only consider setbacks specifically for systems that "
            "would typically be defined as {tech} based on the text itself "
            "— for example, systems intended for electricity generation or "
            "sale, or those above thresholds such as height or rated "
            "capacity. Ignore any requirements that apply only to smaller "
            "or clearly non-commercial systems. "
            "If your answer is 'yes', justify it by quoting the raw text "
            "directly."
        ),
    )
    G.add_edge("non_part", "final")
    G.add_node(
        "final",
        prompt=(
            "Please respond based on our entire conversation so far. "
            "Return your answer as a single dictionary in JSON format (not "
            "markdown). Your JSON file must include exactly two keys. The "
            "keys are 'participating' and 'non-participating'. The value of "
            "the 'participating' key should be a string containing the raw "
            "text with original formatting from the ordinance that applies to "
            "participating owners if you answered 'yes' to the first question "
            "or `null` if you answered 'no'. The value of the "
            "'non-participating' key should be a string "
            "containing the raw text with original formatting from the "
            "ordinance that applies to non-participating owners or simply the "
            "full ordinance for {feature} setbacks if the text did not make "
            "the distinction between participating and non-participating "
            "owners."
        ),
    )
    return G


def setup_graph_extra_restriction(is_numerical=True, **kwargs):
    """Setup Graph to extract non-setback ordinance values from text

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    kwargs.setdefault("unit_clarification", "")
    kwargs.setdefault("feature_clarifications", "")
    G = setup_graph_no_nodes(**kwargs)  # noqa: N806

    G.add_node(
        "init",
        prompt=(
            "Does the following legal text explicitly enact {restriction} for "
            "{tech} for a particular jurisdiction that an energy system "
            "developer would have to abide to? {feature_clarifications}\n"
            "Make sure your answer adheres to these guidelines:\n"
            "1) Respond based only on the explicit text provided for "
            "{restriction}. Do not infer values from text based on "
            "related restrictions. If the text does not explicitly detail "
            "actionable {restriction} for {tech}, please respond with 'No'.\n"
            "2) If the text only defines {restriction} without providing "
            "any specifics, please respond with a simple 'No'.\n"
            "3) Focus only on {restriction} specifically for systems that "
            "would typically be defined as {tech} based on the text itself. "
            "Ignore any requirements that apply only to smaller or clearly "
            "non-commercial systems. "
            "4) Pay close attention to clarifying details in parentheses, "
            "footnotes, or additional explanatory text.\n"
            "5) Please start your response with either 'Yes' or 'No' and "
            "briefly explain your answer."
            '\n\n"""\n{text}\n"""'
        ),
    )

    if is_numerical:
        G.add_edge("init", "value", condition=llm_response_starts_with_yes)
        G.add_node(
            "value",
            prompt=(
                "What is the **numerical** value given for the {restriction} "
                "for {tech}? Follow these guidelines:\n"
                "1) Extract only the explicit numerical value provided for "
                "the restriction. Do not infer values from related "
                "restrictions.\n"
                "2) If multiple values are given, select the most restrictive "
                "one (i.e., the smallest allowable limit, the lowest maximum, "
                "etc.).\n"
                "3) Focus only on {restriction} specifically for systems that "
                "would typically be defined as {tech} based on the text "
                "itself. Ignore any requirements that apply only to smaller "
                "or clearly non-commercial systems. "
                "4) Pay close attention to clarifying details in parentheses, "
                "footnotes, or additional explanatory text.\n\n"
                "Example Inputs and Outputs:\n"
                "Text: 'For all WES there is a limitation of overall height "
                "of 200 feet (including blades).'\n"
                "Output: 200\n"
                "Text: 'The noise level of all SES shall be no greater than "
                "thirty-two (32) decibels measured from the nearest property "
                "line. This level may only be exceeded during short-term "
                "events such as utility outages and/or severe wind storms.'\n"
                "Output: 32\n"
                "Text: 'At no time shall a wind turbine tower, nacelle, or "
                "blade create shadow flicker on any non-participating "
                "landowner property'\n"
                "Output: 0\n"
                "Text: Solar Panels shall not exceed 22'6\" in height. The "
                "height is determined from the ground to the top of the panel "
                "at any angle.\n"
                "Output: 22.5\n"
            ),
        )

        G.add_edge("value", "units")
        G.add_node(
            "units",
            prompt=(
                "What are the units for the {restriction} for {tech}? Ensure "
                "that:\n"
                "1) You accurately identify the unit value associated with "
                "the restriction.\n"
                "2) The unit is expressed using standard, conventional unit "
                "names (e.g., 'feet', 'meters', 'acres', 'dBA', etc.). "
                "{unit_clarification}\n"
                "3) If multiple values are mentioned, return only the units "
                "for the most restrictive value that directly pertains to the "
                "restriction.\n"
                "\nExample Inputs and Outputs:\n"
                "Text: 'For all WES there is a limitation of overall height "
                "of 200 feet (including blades).'\n"
                "Output: 'feet'\n"
                "Text: 'The noise level of all SES shall be no greater than "
                "thirty-two (32) decibels measured from the nearest property "
                "line. This level may only be exceeded during short-term "
                "events such as utility outages and/or severe wind storms.'\n"
                "Output: 'dBA'\n"
                "Text: 'At no time shall a wind turbine tower, nacelle, or "
                "blade create shadow flicker on any non-participating "
                "landowner property'\n"
                "Output: 'hr/year'\n"
            ),
        )

        G.add_edge("units", "final")
        G.add_node(
            "final",
            prompt=(
                "Please respond based on our entire conversation so far. "
                "Return your answer as a dictionary in "
                "JSON format (not markdown). Your JSON file must include "
                "exactly four keys. The keys are 'value', 'units', 'summary', "
                "and 'section'. The value of the '[value]' key "
                "should be a numerical value corresponding to the "
                "{restriction} for {tech}, or `null` if the text "
                "does not mention such a restriction. Use our conversation to "
                "fill out this value. The value of the 'units' key should be "
                "a string corresponding to the (standard) units for the "
                "{restriction} allowed for {tech} by the text "
                "below, or `null` if the text does not mention such a "
                "restriction. "
                "As before, focus only on {restriction} specifically for "
                "systems that would typically be defined as {tech} based on "
                "the text itself. "
                "{SUMMARY_PROMPT} {UNITS_IN_SUMMARY_PROMPT} {SECTION_PROMPT}"
            ),
        )
    else:
        if "moratorium" in kwargs.get("restriction", ""):
            G.add_edge(
                "init",
                "is_conditional",
                condition=llm_response_starts_with_yes,
            )
            G.add_node(
                "is_conditional",
                prompt=(
                    "Does the prohibition, moratorium, or ban only apply "
                    "conditionally? For example, does it only apply to those "
                    "who have not complied with the provisions in this text? "
                    "Please start your response with either 'Yes' or 'No' "
                    "and briefly explain your answer."
                ),
            )
            G.add_edge(
                "is_conditional",
                "has_end_date",
                condition=llm_response_starts_with_no,
            )
            G.add_node(
                "has_end_date",
                prompt=(
                    "Does the legal text given an expiration date for the "
                    "prohibition, moratorium, or ban? "
                    "Please start your response with either 'Yes' or 'No' "
                    "and briefly explain your answer."
                ),
            )
            G.add_edge(
                "has_end_date", "final", condition=llm_response_starts_with_no
            )
            G.add_edge(
                "has_end_date",
                "check_end_date",
                condition=llm_response_starts_with_yes,
            )
            todays_date = datetime.now().strftime("%B %d, %Y")
            G.add_node(
                "check_end_date",
                prompt=(
                    f"Today is {todays_date}. Has the prohibition, "
                    "moratorium, or ban expired? "
                    "Please start your response with either 'Yes' or 'No' "
                    "and briefly explain your answer."
                ),
            )
            G.add_edge(
                "check_end_date",
                "final",
                condition=llm_response_starts_with_no,
            )
        else:
            G.add_edge("init", "final", condition=llm_response_starts_with_yes)
        G.add_node(
            "final",
            prompt=(
                "Please respond based on our entire conversation so far. "
                "Return your answer as a dictionary in "
                "JSON format (not markdown). Your JSON file must include "
                "exactly two keys. The keys are 'summary' and 'section'. "
                "{SUMMARY_PROMPT} {SECTION_PROMPT}"
            ),
        )
    return G


def setup_graph_permitted_use_districts(**kwargs):
    """Setup graph to extract permitted use districts for technology

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = setup_graph_no_nodes(**kwargs)  # noqa: N806

    G.add_node(
        "init",
        prompt=(
            "Does the following legal text explicitly define districts where "
            "{tech} (or similar) are permitted as {use_type}? {clarifications}"
            "Pay extra attention to clarifying text found in "
            "parentheses and footnotes. Please start your response with "
            "either 'Yes' or 'No' and briefly explain your answer."
            '\n\n"""\n{text}\n"""'
        ),
    )
    G.add_edge(
        "init", "district_names", condition=llm_response_starts_with_yes
    )

    G.add_node(
        "district_names",
        prompt=(
            "What are all of the district names (and abbreviations if given) "
            "where {tech} (or similar) are permitted as {use_type}?"
        ),
    )
    G.add_edge("district_names", "final")

    G.add_node(
        "final",
        prompt=(
            "Please respond based on our entire conversation so far. "
            "Return your answer as a dictionary in "
            "JSON format (not markdown). Your JSON file must include "
            "exactly three keys. The keys are 'value', 'summary', "
            "and 'section'. The value of the 'value' key "
            "should be a list of all district names (and abbreviations if "
            "given) where {tech} (or similar) "
            "are permitted as {use_type}, or `null` if the text does not "
            "mention this use type for {tech} (or similar). Use our "
            "conversation to fill out this value. {SUMMARY_PROMPT} "
            "{SECTION_PROMPT}"
        ),
    )
    return G


class BaseTextExtractor:
    """Base implementation for a text extractor"""

    SYSTEM_MESSAGE = (
        "You extract exact excerpts from a provided legal zoning "
        "regulation text that are directly relevant to the user's request "
        "along with relevant context (such as section headers, for example). "
        "You must preserve all original wording, punctuation, and formatting "
        "without any paraphrasing or modification. If the relevant content "
        "appears within a space-delimited table, return the entire table "
        "exactly as formatted. You must not summarize, rephrase, interpret, "
        "or provide additional commentary. Only return the original text "
        "excerpts as they appear in the source."
    )
    _USAGE_LABEL = LLMUsageCategory.DOCUMENT_ORDINANCE_SUMMARY

    def __init__(self, llm_caller):
        """

        Parameters
        ----------
        llm_caller : compass.llm.LLMCaller
            LLM Caller instance used to extract ordinance info with.
        """
        self.llm_caller = llm_caller

    async def _process(self, text_chunks, instructions, is_valid_chunk):
        """Perform extraction processing"""
        logger.info(
            "Extracting summary text from %d text chunks asynchronously...",
            len(text_chunks),
        )
        logger.debug("Model instructions are:\n%s", instructions)
        outer_task_name = asyncio.current_task().get_name()
        summaries = [
            asyncio.create_task(
                self.llm_caller.call(
                    sys_msg=self.SYSTEM_MESSAGE,
                    content=f"{instructions}\n\n# TEXT #\n\n{chunk}",
                    usage_sub_label=self._USAGE_LABEL,
                ),
                name=outer_task_name,
            )
            for chunk in text_chunks
        ]
        summary_chunks = await asyncio.gather(*summaries)
        summary_chunks = [
            clean_backticks_from_llm_response(chunk)
            for chunk in summary_chunks
            if is_valid_chunk(chunk)
        ]

        text_summary = merge_overlapping_texts(summary_chunks)
        logger.debug(
            "Final summary contains %d tokens",
            ApiBase.count_tokens(
                text_summary,
                model=self.llm_caller.kwargs.get("model", "gpt-4"),
            ),
        )
        return text_summary



"""Ordinance jurisdiction info"""

import logging
from warnings import warn
from pathlib import Path

import numpy as np
import pandas as pd

from compass.exceptions import COMPASSValueError
from compass.warn import COMPASSWarning


logger = logging.getLogger(__name__)
_COUNTY_DATA_FP = (
    Path(__file__).parent.parent / "data" / "conus_jurisdictions.csv"
)


def load_all_jurisdiction_info():
    """Load DataFrame containing info for all jurisdictions

    Returns
    -------
    pd.DataFrame
        DataFrame containing info like names, FIPS, websites, etc. for
        all jurisdictions.
    """
    jurisdiction_info = pd.read_csv(_COUNTY_DATA_FP).replace({np.nan: None})
    jurisdiction_info = _convert_to_title(jurisdiction_info, "State")
    jurisdiction_info = _convert_to_title(jurisdiction_info, "County")
    return _convert_to_title(jurisdiction_info, "Subdivision")


def jurisdiction_websites(jurisdiction_info=None):
    """Load mapping of jurisdiction name and state to website

    Parameters
    ----------
    jurisdiction_info : pd.DataFrame, optional
        DataFrame containing jurisdiction names and websites. If
        ``None``, this info is loaded using
        :func:`load_jurisdiction_info`.
        By default, ``None``.

    Returns
    -------
    dict
        Dictionary where keys are FIPS codes and values are the relevant
        website URL.
    """
    if jurisdiction_info is None:
        jurisdiction_info = load_all_jurisdiction_info()

    return {
        row["FIPS"]: row["Website"] for __, row in jurisdiction_info.iterrows()
    }


def load_jurisdictions_from_fp(jurisdiction_fp):
    """Load jurisdiction info based on jurisdictions in the input fp

    Parameters
    ----------
    jurisdiction_fp : path-like
        Path to csv file containing "County" and "State" columns that
        define the jurisdictions for which info should be loaded.

    Returns
    -------
    pd.DataFrame
        DataFrame containing jurisdiction info like names, FIPS,
        websites, etc. for all requested jurisdictions (that were
        found).
    """
    jurisdictions = pd.read_csv(jurisdiction_fp)
    jurisdictions = _validate_jurisdiction_input(jurisdictions)

    all_jurisdiction_info = load_all_jurisdiction_info()
    merge_cols = ["County", "State"]
    if "Subdivision" in jurisdictions:
        merge_cols += ["Subdivision", "Jurisdiction Type"]
    else:
        all_jurisdiction_info = all_jurisdiction_info[
            all_jurisdiction_info["Subdivision"].isna()
        ].reset_index(drop=True)

    jurisdictions = jurisdictions.merge(
        all_jurisdiction_info, on=merge_cols, how="left"
    )

    jurisdictions = _filter_not_found_jurisdictions(jurisdictions)
    return _format_jurisdiction_df_for_output(jurisdictions)


def _validate_jurisdiction_input(jurisdictions):
    """Throw error if user is missing required columns"""
    if "State" not in jurisdictions:
        msg = "The jurisdiction input must have at least a 'State' column!"
        raise COMPASSValueError(msg)

    jurisdictions = _convert_to_title(jurisdictions, "State")

    if "County" not in jurisdictions:
        jurisdictions["County"] = None
    else:
        jurisdictions = _convert_to_title(jurisdictions, "County")

    if "Subdivision" in jurisdictions:
        if "Jurisdiction Type" not in jurisdictions:
            msg = (
                "The jurisdiction input must have a 'Jurisdiction Type' "
                "column if a 'Subdivision' column is provided (this helps "
                "avoid name clashes for certain subdivisions)!"
            )
            raise COMPASSValueError(msg)

        jurisdictions = _convert_to_title(jurisdictions, "Subdivision")
        jurisdictions["Jurisdiction Type"] = jurisdictions[
            "Jurisdiction Type"
        ].str.casefold()

    return jurisdictions


def _filter_not_found_jurisdictions(df):
    """Filter out jurisdictions with null FIPS codes"""
    _warn_about_missing_jurisdictions(df)
    return df[~df["FIPS"].isna()].copy()


def _warn_about_missing_jurisdictions(df):
    """Throw warning about jurisdictions that were not in the list"""
    not_found_jurisdictions = df[df["FIPS"].isna()]
    if len(not_found_jurisdictions):
        not_found_jurisdictions_str = not_found_jurisdictions[
            ["State", "County", "Subdivision", "Jurisdiction Type"]
            # cspell: disable-next-line
        ].to_markdown(index=False, tablefmt="psql")
        msg = (
            "The following jurisdictions were not found! Please make sure to "
            "use proper spelling and capitalization.\n"
            f"{not_found_jurisdictions_str}"
        )
        warn(msg, COMPASSWarning)


def _format_jurisdiction_df_for_output(df):
    """Format jurisdiction DataFrame for output"""
    out_cols = [
        "County",
        "State",
        "Subdivision",
        "Jurisdiction Type",
        "FIPS",
        "Website",
    ]
    df["FIPS"] = df["FIPS"].astype(int)
    return df[out_cols].replace({np.nan: None}).reset_index(drop=True)


def _convert_to_title(df, column):
    """Convert the values of a DataFrame column to titles"""
    df[column] = df[column].str.strip().str.casefold().str.title()
    return df



"""COMPASS common extraction utilities and graphs"""

from .base import (
    EXTRACT_ORIGINAL_TEXT_PROMPT,
    BaseTextExtractor,
    empty_output,
    found_ord,
    llm_response_starts_with_no,
    llm_response_starts_with_yes,
    run_async_tree,
    run_async_tree_with_bm,
    setup_async_decision_tree,
    setup_base_setback_graph,
    setup_graph_extra_restriction,
    setup_graph_no_nodes,
    setup_graph_permitted_use_districts,
    setup_participating_owner,
)
